# -*- coding: utf-8 -*-

"""
word2word/methods.py: bilingual lexicon extraction methods

Speed comparison using OpenSubtitles.en-eo (64,485 sentences):
    - rerank (single process): 105s
    - rerank_mp (multiprocessing), 8 CPUs: 33s (3.2x faster)
    - rerank_mp (multiprocessing), 16 CPUs: 26s (4.0x faster)
    - rerank_mp (multiprocessing), 32 CPUs: 47s (2.2x faster)

Optimal number of CPUs may differ depending on corpus size.
"""

import itertools as it
import numpy as np
import operator
from tqdm import tqdm


def rerank(x2ys, x2cnt, x2xs, width, n_trans):
    """Re-rank word translations by computing CPE scores.

    See paper for details about the CPE method."""
    x2ys_cpe = dict()
    for x, ys in tqdm(x2ys.items()):
        cntx = x2cnt[x]
        y_scores = []
        for y, cnty in sorted(ys.items(), key=operator.itemgetter(1), reverse=True)[:width]:
            ts = cnty / float(cntx)  # translation score: initial value
            if x in x2xs:
                for x2, cntx2 in x2xs[x].items():  # Collocates
                    p_x_x2 = cntx2 / float(cntx)
                    p_x2_y2 = 0
                    if x2 in x2ys:
                        p_x2_y2 = x2ys[x2].get(y, 0) / float(x2cnt[x2])
                    ts -= (p_x_x2 * p_x2_y2)
            y_scores.append((y, ts))
        _ys_ = sorted(y_scores, key=lambda y_score: y_score[1], reverse=True)[:n_trans]
        _ys_ = [each[0] for each in _ys_]
        x2ys_cpe[x] = _ys_

    return x2ys_cpe


def _rerank_mp(x_and_ys, shared_inputs):
    """Internal multiprocessing function for `rerank_fast()`."""
    x, ys = x_and_ys
    x2ys, x2cnt, x2xs, width, n_trans = shared_inputs

    sorted_ys = sorted(ys.items(),
                       key=operator.itemgetter(1),
                       reverse=True)[:width]
    if x not in x2xs:
        return x, [y for y, score in sorted_ys[:n_trans]]

    def _correction(y):
        return sum(
            cntx2 * x2ys[x2][y] / float(x2cnt[x2])
            for x2, cntx2 in x2xs[x].items() if x2 in x2ys and y in x2ys[x2]
        )

    y_scores = [(y, cnty - _correction(y)) for y, cnty in sorted_ys]
    y_scores = sorted(y_scores, key=operator.itemgetter(1), reverse=True)
    reranked_ys = [y for y, score in y_scores[:n_trans]]
    return x, reranked_ys


def rerank_mp(x2ys, x2cnt, x2xs, width, n_trans, num_workers):
    """Re-rank word translations by computing CPE scores.

    Uses multiprocessing to speed up computation (significantly).
    In Python 3.8+, shared_inputs can be implemented directly as shared_memory.

    See paper for details about the CPE method."""
    from multiprocessing import Pool

    shared_inputs = x2ys, x2cnt, x2xs, width, n_trans
    print(f"Entering multiprocessing with {num_workers} workers..."
          f" (#words={len(x2ys)})")
    with Pool(num_workers) as p:
        x2ys_cpe = dict(p.starmap(
            _rerank_mp,
            zip(x2ys.items(), it.repeat(shared_inputs)),
        ))
    return x2ys_cpe


def get_trans_co(x2ys, n_trans):
    """Use co-occurrences to compute scores."""
    x2ys_co = dict()
    for x, ys in x2ys.items():
        ys = [y for y, cnt in sorted(ys.items(), key=operator.itemgetter(1), reverse=True)[:n_trans]]
        x2ys_co[x] = ys
    return x2ys_co


def get_trans_pmi(x2ys, x2cnt, y2cnt, Nxy, Nx, Ny, width, n_trans):
    """Use pointwise mutual information to compute scores."""
    x2ys_pmi = dict()
    pmi_diff = -np.log2(Nxy) + np.log2(Nx) + np.log2(Ny)
    for x, ys in tqdm(x2ys.items()):
        l_scores = []
        for y, cnt in sorted(ys.items(), key=operator.itemgetter(1),
                             reverse=True)[:width]:
            pmi = np.log2(cnt) - np.log2(x2cnt[x]) - np.log2(y2cnt[y])
            pmi += pmi_diff
            l_scores.append((y, pmi))
        trans = sorted(l_scores, key=operator.itemgetter(1, 0), reverse=True)[:n_trans]
        trans = [each[0] for each in trans]
        x2ys_pmi[x] = trans

    return x2ys_pmi



# -*- coding: utf-8 -*-

import os
import pickle
from time import time

from word2word.utils import (
    download_or_load, download_os2018, get_savedir
)
from word2word.tokenization import (
    load_tokenizer, get_sents, get_vocab, update_dicts
)
from word2word.methods import (
    rerank, rerank_mp, get_trans_co, get_trans_pmi
)


class Word2word:
    """The word2word class.

    Usage:
        from word2word import Word2word

        # Download and load a pre-computed bilingual lexicon
        en2fr = Word2word("en", "fr")
        print(en2fr("apple"))
        # out: ['pomme', 'pommes', 'pommier', 'tartes', 'fleurs']

        # Build a custom bilingual lexicon
        # (requires two aligned files, e.g., my_corpus.en, my_corpus.fr)
        my_en2fr = Word2word.make("en", "fr", "my_corpus")
    """

    def __init__(self, lang1, lang2,
                 word2x=None, y2word=None, x2ys=None, custom_savedir=None):
        self.lang1 = lang1
        self.lang2 = lang2

        if all(d is not None for d in [word2x, y2word, x2ys]):
            # load a custom-built word2word bilingual lexicon
            self.word2x, self.y2word, self.x2ys = word2x, y2word, x2ys
        elif any(d is not None for d in [word2x, y2word, x2ys]):
            raise ValueError(
                f"custom bilingual lexicon is only partially provided. "
                f"use Word2word.make() or Word2word.load() to "
                f"properly build or load custom lexicons.")
        else:
            # default: download/load the word2word dataset
            self.word2x, self.y2word, self.x2ys = download_or_load(
                lang1, lang2, custom_savedir
            )

        # lazily keep summary statistics of the learned bilingual lexicon
        self.summary = {}

    def __call__(self, query, n_best=5):
        """Retrieve top-k word translations for the query word."""
        try:
            x = self.word2x[query]
            ys = self.x2ys[x]
            words = [self.y2word[y] for y in ys]
        except KeyError:
            raise KeyError(
                f"query word {query} not found in the bilingual lexicon."
            )
        return words[:n_best]

    def __len__(self):
        """Return the number of source words for which translation exists."""
        return len(self.x2ys)

    def compute_summary(self):
        """Compute basic summaries for the bilingual lexicon."""
        n_unique_ys = len(set([y for ys in self.x2ys.values() for y in ys]))
        n_ys = [len(ys) for ys in self.x2ys.values()]
        self.summary = {
            "n_valid_words": len(self),
            "n_valid_targets": n_unique_ys,
            "n_total_words": len(self.word2x),
            "n_total_targets": len(self.y2word),
            "n_translations_per_word": sum(n_ys) / len(n_ys),
            "n_sentences": None,  # original file required
        }
        return self.summary

    @classmethod
    def make(
            cls,
            lang1: str,
            lang2: str,
            datapref: str = None,
            n_lines: int = 100000000,
            cutoff: int = 5000,
            rerank_width: int = 100,
            rerank_impl: str = "multiprocessing",
            cased: bool = True,
            n_translations: int = 10,
            save_cooccurrence: bool = False,
            save_pmi: bool = False,
            savedir: str = None,
            num_workers: int = 16,
    ):
        """Build a bilingual lexicon using a parallel corpus."""

        print("Step 0. Check files")
        lang1, lang2 = sorted([lang1, lang2])
        if datapref:
            lang1_file, lang2_file = [
                f"{datapref}.{lang}" for lang in [lang1, lang2]
            ]
            assert os.path.exists(lang1_file), \
                f"custom parallel corpus file missing at {datapref}.{lang1}"
            assert os.path.exists(lang2_file), \
                f"custom parallel corpus file missing at {datapref}.{lang2}"
        else:
            lang1_file, lang2_file = download_os2018(lang1, lang2)

        print("Step 1. Load tokenizer")
        tokenizer1 = load_tokenizer(lang1)
        tokenizer2 = load_tokenizer(lang2)

        t0 = time()
        print("Step 2. Constructing sentences")
        sents1 = get_sents(
            lang1_file, lang1, tokenizer1, cased, n_lines, num_workers
        )
        sents2 = get_sents(
            lang2_file, lang2, tokenizer2, cased, n_lines, num_workers
        )
        print(f"Time taken for step 2: {time() - t0:.2f}s")

        assert len(sents1) == len(sents2), (
            f"{lang1} and {lang2} files must have the same number of lines.\n"
            f"({lang1}: {len(sents1)} lines, {lang2}: {len(sents2)} lines)"
        )

        # input savedir if provided, else datapref (custom data location);
        # system default otherwise
        savedir = get_savedir(savedir if savedir else datapref)

        print("Step 3. Compute vocabularies")
        # word <-> index
        word2x, x2word, x2cnt = get_vocab(sents1)
        word2y, y2word, y2cnt = get_vocab(sents2)

        print("Step 4. Update count dictionaries")
        # monolingual and cross-lingual dictionaries
        x2xs, y2ys, x2ys, y2xs = update_dicts(
            sents1, sents2, word2x, word2y, cutoff
        )

        t0 = time()
        print("Step 5. Translation using CPE scores")
        if rerank_impl == "simple":
            x2ys_cpe = rerank(x2ys, x2cnt, x2xs, rerank_width, n_translations)
            y2xs_cpe = rerank(y2xs, y2cnt, y2ys, rerank_width, n_translations)
        elif rerank_impl == "multiprocessing":
            x2ys_cpe = rerank_mp(
                x2ys, x2cnt, x2xs, rerank_width, n_translations, num_workers
            )
            y2xs_cpe = rerank_mp(
                y2xs, y2cnt, y2ys, rerank_width, n_translations, num_workers
            )
        else:
            raise ValueError("unrecognized --rerank_impl argument. "
                             "Options: simple, multiprocessing")
        print(f"Time taken for step 5: {time() - t0:.2f}s")

        print("Saving...")
        Word2word.save(lang1, lang2, savedir, word2x, word2y, x2word,
                       x2ys_cpe, y2word, y2xs_cpe)

        if save_cooccurrence:
            print("Step 5-1. Translation using co-occurrence counts")
            subdir = os.path.join(savedir, "co")
            os.makedirs(subdir, exist_ok=True)

            x2ys_co = get_trans_co(x2ys, n_translations)
            y2xs_co = get_trans_co(y2xs, n_translations)
            Word2word.save(lang1, lang2, subdir, word2x, word2y, x2word,
                           x2ys_co, y2word, y2xs_co)

        if save_pmi:
            print("Step 5-2. Translation using PMI scores")
            subdir = os.path.join(savedir, "pmi")
            os.makedirs(subdir, exist_ok=True)

            seqlens1 = [len(sent) for sent in sents1]
            seqlens2 = [len(sent) for sent in sents2]
            Nx = sum(seqlens1)
            Ny = sum(seqlens2)
            Nxy = sum([seqlen_x * seqlen_y
                       for seqlen_x, seqlen_y in zip(seqlens1, seqlens2)])

            x2ys_pmi = get_trans_pmi(x2ys, x2cnt, y2cnt, Nxy, Nx, Ny,
                                     rerank_width, n_translations)
            y2xs_pmi = get_trans_pmi(y2xs, y2cnt, x2cnt, Nxy, Ny, Nx,
                                     rerank_width, n_translations)

            Word2word.save(lang1, lang2, subdir, word2x, word2y, x2word,
                           x2ys_pmi, y2word, y2xs_pmi)

        print("Done!")
        return cls(lang1, lang2, word2x, y2word, x2ys_cpe)

    @staticmethod
    def save(lang1, lang2, savedir, word2x, word2y, x2word, x2ys, y2word, y2xs):
        with open(os.path.join(savedir, f"{lang1}-{lang2}.pkl"), "wb") as f:
            pickle.dump((word2x, y2word, x2ys), f)
        with open(os.path.join(savedir, f"{lang2}-{lang1}.pkl"), "wb") as f:
            pickle.dump((word2y, x2word, y2xs), f)

    @classmethod
    def load(cls, lang1, lang2, savedir):
        """Loads this object with a custom-built bilingual lexicon.

        savedir is the directory containing {lang1}-{lang2}.pkl files
        built from the make function.
        """
        path = os.path.join(savedir, f"{lang1}-{lang2}.pkl")
        assert os.path.exists(path), \
            f"processed lexicon file not found at {path}"
        with open(path, "rb") as f:
            word2x, y2word, x2ys = pickle.load(f)
        print(f"Loaded word2word custom bilingual lexicon from {path}")
        return cls(lang1, lang2, word2x, y2word, x2ys)



# -*- coding: utf-8 -*-
import platform
import wget
import requests
import os
import pickle
from zipfile import ZipFile
from os.path import expanduser


def get_savedir(savedir=None):
    if savedir:
        os.makedirs(savedir, exist_ok=True)
        return savedir

    pf = platform.system()
    if pf == "Windows":
        savedir = "C:\word2word"
    else:
        homedir = expanduser("~")
        savedir = os.path.join(homedir, ".word2word")

    if not os.path.exists(savedir):
        os.makedirs(savedir, exist_ok=True)
    return savedir


def exists(path):
    r = requests.head(path)
    return r.status_code == requests.codes.ok


def get_download_url(lang1, lang2):
    filepath = os.path.dirname(os.path.abspath(__file__)) + '/supporting_languages.txt'
    for line in open(filepath, 'r'):
        l1, l2 = line.strip().split("-")
        if lang1 == l1 and lang2 == l2:
            return f"https://mk.kakaocdn.net/dn/kakaobrain/word2word/{lang1}-{lang2}.pkl"
    raise Exception(f"Language pair {lang1}-{lang2} is not supported.")


def download_or_load(lang1, lang2, custom_savedir):
    savedir = get_savedir(savedir=custom_savedir)
    fpath = os.path.join(savedir, f"{lang1}-{lang2}.pkl")
    if not os.path.exists(fpath):
        # download from cloud
        url = get_download_url(lang1, lang2)
        if url is None:
            raise ValueError(f"Dataset not found for {lang1}-{lang2}.")

        if not exists(url):
            raise ValueError("Sorry. There seems to be a problem with cloud access.")

        print("Downloading data ...")
        wget.download(url, fpath)
    word2x, y2word, x2ys = pickle.load(open(fpath, 'rb'))
    return word2x, y2word, x2ys


def download_os2018(lang1, lang2):
    """Download corpora from OpenSubtitles2018.

    :return (lang1_file, lang2_file)
    """
    datadir = "data"
    filepref = f"OpenSubtitles.{lang1}-{lang2}"
    if all(os.path.exists(os.path.join(datadir, f"{filepref}.{lang}"))
           for lang in [lang1, lang2]):
        print(f"Found existing {filepref} files. loading...")
    else:
        # Download and unzip parallel corpus
        url = f"http://opus.nlpl.eu/download.php?f=OpenSubtitles/v2018/moses/{lang1}-{lang2}.txt.zip"
        zipname = os.path.join(datadir, f"{lang1}-{lang2}.txt.zip")
        print(f"Downloading {filepref}...")
        wget.download(url, zipname)
        with ZipFile(zipname) as zf:
            for fname in zf.namelist():
                if fname.startswith(filepref):
                    zf.extract(fname, datadir)
        os.remove(zipname)
    lang1_file, lang2_file = [
        os.path.abspath(os.path.join(datadir, f"{filepref}.{lang}"))
        for lang in [lang1, lang2]
    ]
    return lang1_file, lang2_file



import setuptools

with open("README.md", mode="r", encoding="utf-8") as fh:
    long_description = fh.read()

REQUIRED_PACKAGES = [
    'requests',
    'wget',
    'numpy',
    'tqdm',
]

setuptools.setup(
    name="word2word",
    version="1.0.0",
    author="Kyubyong Park, Dongwoo Kim, Yo Joong Choe",
    author_email="kbpark.linguist@gmail.com, kimdwkimdw@gmail.com, yjchoe33@gmail.com",
    description="Easy-to-use word translations for 3,564 language pairs",
    install_requires=REQUIRED_PACKAGES,
    license='Apache License 2.0',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/kakaobrain/word2word",
    packages=setuptools.find_packages(),
    package_data={'word2word': ['word2word/supporting_languages.txt']},
    python_requires=">=3.6",
    include_package_data=True,
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
)



# -*- coding: utf-8 -*-

from collections import Counter, defaultdict
from itertools import chain, islice, product, repeat
from multiprocessing import Pool
import operator
import re
from tqdm import tqdm


def load_tokenizer(lang):
    if lang == "ko":
        from konlpy.tag import Mecab
        tokenizer = Mecab()
    elif lang == "ja":
        import Mykytea
        opt = "-model jp-0.4.7-1.mod"
        tokenizer = Mykytea.Mykytea(opt)
    elif lang == "zh_cn":
        import Mykytea
        opt = "-model ctb-0.4.0-1.mod"
        tokenizer = Mykytea.Mykytea(opt)
    elif lang == "zh_tw":
        import jieba
        tokenizer = jieba
    elif lang == "vi":
        from pyvi import ViTokenizer
        tokenizer = ViTokenizer
    elif lang == "th":
        from pythainlp.tokenize import word_tokenize
        tokenizer = word_tokenize
    elif lang == "ar":
        import pyarabic.araby as araby
        tokenizer = araby
    # elif lang=="en":
    #     from nltk import word_tokenize
    #     tokenizer = word_tokenize
    else:
        from nltk.tokenize import ToktokTokenizer
        tokenizer = ToktokTokenizer()

    return tokenizer


def word_segment(sent, lang, tokenizer):
    if lang == 'ko':
        words = [word for word, _ in tokenizer.pos(sent)]
    elif lang == 'ja':
        words = [elem for elem in tokenizer.getWS(sent)]
    elif lang == 'th':
        words = tokenizer(sent, engine='mm')
    elif lang == 'vi':
        words = tokenizer.tokenize(sent).split()
    elif lang == 'zh_cn':
        words = [elem for elem in tokenizer.getWS(sent)]
    elif lang == "zh_tw":
        words = list(tokenizer.cut(sent, cut_all=False))
    elif lang == "ar":
        words = tokenizer.tokenize(sent)
    # elif lang=="en":
    #     words = tokenizer(sent)
    else:  # Most european languages
        sent = re.sub("([A-Za-z])(\.[ .])", r"\1 \2", sent)
        words = tokenizer.tokenize(sent)

    return words


def process_line(line, lang, tokenizer, cased):
    """Strip, uncase (optionally), and tokenize line.

    multiprocessing helper for get_sents()."""
    line = line.strip() if cased else line.strip().lower()
    return word_segment(line, lang, tokenizer)


def get_sents(fin, lang, tokenizer, cased, n_lines, num_workers=8):
    """Load parallel corpus and segment words using multiprocessing."""

    with open(fin, encoding='utf-8') as f:
        lines = islice(f, n_lines)
        if num_workers <= 1:
            return [process_line(line, lang, tokenizer, cased)
                    for line in lines]
        else:
            print(f"Entering multiprocessing with {num_workers} workers...")
            with Pool(num_workers) as p:
                return p.starmap(
                    process_line,
                    zip(lines, repeat(lang), repeat(tokenizer), repeat(cased))
                )


def get_vocab(sents):
    word2idx, idx2word, idx2cnt = dict(), dict(), dict()

    word2cnt = Counter(tqdm(list(chain.from_iterable(sents)))).most_common()
    word2cnt.sort(key=operator.itemgetter(1, 0), reverse=True)
    for idx, (word, cnt) in enumerate(tqdm(word2cnt)):
        word2idx[word] = idx
        idx2word[idx] = word
        idx2cnt[idx] = cnt

    return word2idx, idx2word, idx2cnt


def update_dicts(sents1, sents2, vocab1, vocab2, cutoff):
    """Get monolingual and cross-lingual count dictionaries.

    'cutoff' determines how many collocates are considered in each language.
    """

    def u2_iter(t1, t2, same_ignore=False, cut_t2=None):
        for _ in product(t1, t2):
            if (not same_ignore or _[0] != _[1]) and (not cut_t2 or _[1] < cut_t2):
                yield _

    def build_ddi():
        return defaultdict(lambda: defaultdict(int))

    x_x_dict = build_ddi()
    y_y_dict = build_ddi()
    x_y_dict = build_ddi()
    y_x_dict = build_ddi()

    for sent1, sent2 in tqdm(zip(sents1, sents2), total=len(sents1)):
        xs = [vocab1[wx] for wx in sent1 if wx in vocab1]
        ys = [vocab2[wy] for wy in sent2 if wy in vocab2]

        for xx1, xx2 in u2_iter(xs, xs, same_ignore=True, cut_t2=cutoff):
            x_x_dict[xx1][xx2] += 1
        for yy1, yy2 in u2_iter(ys, ys, same_ignore=True, cut_t2=cutoff):
            y_y_dict[yy1][yy2] += 1
        for xx, yy in u2_iter(xs, ys, same_ignore=False):
            x_y_dict[xx][yy] += 1
            y_x_dict[yy][xx] = x_y_dict[xx][yy]

    # convert to ordinary dicts for pickling
    def ddi2dict(ddi):
        return {k: dict(v) for k, v in ddi.items()}

    return tuple(ddi2dict(ddi)
                 for ddi in [x_x_dict, y_y_dict, x_y_dict, y_x_dict])
